import 'package:boost/boost.dart';
import 'package:cl_appbase/appbase.dart';
import 'package:flutter/material.dart';

class TableCollectionBuilder<Item> extends StatefulWidget {
  final ReactiveCollection<Item> collection;
  final List<DataColumn> columns;
  final TableRowBuilder<Item> rowBuilder;
  final WidgetBuilder? loadingBuilder;
  final ErrorBuilder? errorBuilder;
  final double? columnSpacing;
  final bool pullToRefresh;
  final ScrollPhysics? physics;

  const TableCollectionBuilder({
    Key? key,
    required this.collection,
    required this.columns,
    required this.rowBuilder,
    this.loadingBuilder,
    this.errorBuilder,
    this.columnSpacing,
    this.pullToRefresh = false,
    this.physics,
  }) : super(key: key);

  @override
  _TableCollectionBuilderState<Item> createState() =>
      _TableCollectionBuilderState<Item>();
}

class _TableCollectionBuilderState<Item>
    extends State<TableCollectionBuilder<Item>> {
  final _continueSemaphore = Semaphore();
  bool _done = false;

  @override
  void initState() {
    super.initState();
    _continueLoading();
  }

  @override
  Widget build(BuildContext context) {
    return CollectionBuilder(
      stream: widget.collection.stream,
      builder: _build,
      error: _buildError,
    );
  }

  Widget _build(BuildContext context, List<Item> items, bool loading) {
    if (widget.pullToRefresh) {
      return RefreshIndicator(
        child: _buildScrollView(context, items, loading),
        onRefresh: _refresh,
      );
    } else {
      return _buildScrollView(context, items, loading);
    }
  }

  Widget _buildScrollView(
      BuildContext context, List<Item> items, bool loading) {
    return NotificationListener<ScrollNotification>(
      onNotification: _onScroll,
      child: SingleChildScrollView(
        physics: widget.physics,
        child: Column(
          crossAxisAlignment: CrossAxisAlignment.stretch,
          mainAxisSize: MainAxisSize.min,
          children: [
            _buildTable(context, items),
            if (loading)
              Center(
                child: _buildLoading(context),
              ),
          ],
        ),
      ),
    );
  }

  Widget _buildLoading(BuildContext context) {
    if (widget.loadingBuilder != null) {
      return widget.loadingBuilder!(context);
    } else {
      // fallback loading indicator
      return ConstrainedBox(
        constraints: const BoxConstraints(
          maxHeight: 16,
          maxWidth: 16,
        ),
        child: const CircularProgressIndicator(),
      );
    }
  }

  Widget _buildTable(BuildContext context, List<Item> items) {
    return DataTable(
      columnSpacing: widget.columnSpacing,
      columns: widget.columns,
      rows: items
          .map((item) => widget.rowBuilder(context, item))
          .toList(growable: false),
    );
  }

  Widget _buildError(BuildContext context, dynamic error) {
    if (widget.loadingBuilder != null) {
      return widget.loadingBuilder!(context);
    } else {
      // fallback loading indicator
      return Text(error.toString()); //TODO errorText
    }
  }

  bool _onScroll(ScrollNotification notification) {
    if (!_done &&
        notification.metrics.pixels >
            notification.metrics.maxScrollExtent - 300) {
      _continueLoading();
    }
    return false; //don't intercept bubbling
  }

  Future<void> _continueLoading({bool refresh = false}) async {
    await _continueSemaphore.throttle(() async {
      _done = await widget.collection.load(invalidate: refresh);
    });
  }

  Future<void> _refresh() async {
    await _continueLoading(refresh: true);
  }
}
